@@ -6,24 +6,23 @@ __author__ = 'Jens-Kristian Krogager'
 __email__ = "krogager@iap.fr"
 __credits__ = ["Jens-Kristian Krogager"]
 
-from argparse import ArgumentParser
 import numpy as np
 import astropy.io.fits as pf
 import matplotlib.pyplot as plt
 from matplotlib import ticker
-from scipy.ndimage import gaussian_filter1d
-from numpy.polynomial import Chebyshev
+from scipy import ndimage, signal
 import os
 from os.path import exists, basename
 
 from pynot import alfosc
 from pynot.functions import mad, my_formatter, get_version_number
+from pynot.scired import trim_overscan
 
 
 __version__ = get_version_number()
 
 
-def combine_bias_frames(bias_frames, output='', kappa=15, overwrite=True):
+def combine_bias_frames(bias_frames, output='', kappa=15, overwrite=True, overscan=50, mode='spec'):
     """Combine individual bias frames to create a 'master bias' frame.
     The combination is performed using robust sigma-clipping and
     median combination. Bad pixels are subsequently replaced by the
@@ -58,9 +57,15 @@ def combine_bias_frames(bias_frames, output='', kappa=15, overwrite=True):
     for frame in bias_frames:
         msg.append("          - Loaded bias frame: %s" % frame)
         raw_img = pf.getdata(frame)
+        bias_hdr = alfosc.get_alfosc_header(frame)
+        if mode == 'spec':
+            trim_bias, bias_hdr = trim_overscan(raw_img, bias_hdr, overscan)
+            msg.append("          - Trimming overscan of bias images: %i!" % overscan)
+        else:
+            trim_bias = raw_img
         if len(bias) > 1:
-            assert raw_img.shape == bias[-1].shape, "Images must have same shape!"
-        bias.append(raw_img)
+            assert trim_bias.shape == bias[0].shape, "Images must have same shape!"
+        bias.append(trim_bias)
 
     mask = np.zeros_like(bias[0], dtype=int)
     median_img0 = np.median(bias, 0)
@@ -98,7 +103,7 @@ def combine_bias_frames(bias_frames, output='', kappa=15, overwrite=True):
 
 
 def combine_flat_frames(raw_frames, output, mbias='', mode='spec', dispaxis=2,
-                        kappa=5, verbose=False, overwrite=True):
+                        kappa=5, verbose=False, overwrite=True, overscan=50):
     """Combine individual spectral flat frames to create a 'master flat' frame.
     The individual frames are normalized to the mode of the 1D collapsed spectral
     shape. Individual frames are clipped using a kappa-sigma-clipping on the mode
@@ -148,6 +153,10 @@ def combine_flat_frames(raw_frames, output, mbias='', mode='spec', dispaxis=2,
     msg = list()
     if mbias and exists(mbias):
         bias = pf.getdata(mbias)
+        bias_hdr = alfosc.get_alfosc_header(mbias)
+        if mode == 'spec':
+            bias, bias_hdr = trim_overscan(bias, bias_hdr, overscan)
+            msg.append("          - Trimming overscan of bias image: %i!" % overscan)
     else:
         msg.append("[WARNING] - No master bias frame provided!")
         bias = 0.
@@ -155,17 +164,21 @@ def combine_flat_frames(raw_frames, output, mbias='', mode='spec', dispaxis=2,
     flats = list()
     flat_peaks = list()
     for fname in raw_frames:
-        hdr = pf.getheader(fname)
+        hdr = alfosc.get_alfosc_header(fname)
         flat = pf.getdata(fname)
-        flat = flat - bias
         if mode == 'spec':
+            flat, hdr = trim_overscan(flat, hdr, overscan)
+            flat = flat - bias
             peak_val = np.max(np.mean(flat, dispaxis-1))
             flats.append(flat/peak_val)
             flat_peaks.append(peak_val)
             msg.append("          - Loaded Spectral Flat file: %s   mode=%.1f" % (fname, peak_val))
+            msg.append("          - Trimming overscan of Flat images: %i!" % overscan)
 
         else:
-            peak_val = np.median(flat, 1)
+            flat = flat - bias
+            pad = np.max(flat.shape) // 4
+            peak_val = np.median(flat[pad:-pad, pad:-pad])
             flats.append(flat/peak_val)
             flat_peaks.append(peak_val)
             msg.append("          - Loaded Imaging Flat file: %s   median=%.1f" % (fname, peak_val))
@@ -179,7 +192,7 @@ def combine_flat_frames(raw_frames, output, mbias='', mode='spec', dispaxis=2,
         masked_flats.append(np.ma.masked_where(this_mask, img))
         mask += 1*this_mask
     msg.append("          - Standard deviation of raw median image: %.1f ADUs" % sig)
-    msg.append("          - Masking outlying pixels using a threshold of  kappa=%.1f" % kappa)
+    msg.append("          - Masking outlying pixels using a threshold of kappa=%.1f" % kappa)
     msg.append("          - Total number of masked pixels: %i" % np.sum(mask > 0))
     msg.append("          - Median value of combined flat: %i" % np.sum(mask > 0))
 
@@ -228,8 +241,27 @@ def combine_flat_frames(raw_frames, output, mbias='', mode='spec', dispaxis=2,
     return output, output_msg
 
 
-def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, lower=0, upper=2050, order=24, sigma=5,
-                            plot=True, show=True, ext=1, overwrite=True, verbose=False):
+def detect_flat_edges(img, dispaxis=2, savgol_window=21, threshold=10):
+    """Use filtered derivative to detect the edges of the frame."""
+    # Get median profile along slit:
+    med_row = np.nanmedian(img, 2-dispaxis)
+
+    # Calculate the absolute value of the derivative:
+    deriv = np.fabs(signal.savgol_filter(med_row, savgol_window, 1, deriv=1))
+
+    # Find peaks of the derivative:
+    noise = 1.48*mad(deriv)
+    if noise == 0:
+        sig_tmp = np.std(deriv)
+        noise = np.std(deriv[deriv < 3*sig_tmp])
+    edges, props = signal.find_peaks(deriv, height=threshold*noise)
+
+    return edges
+
+
+
+def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, overscan=50, order=24, savgol_window=51,
+                            med_window=5, edge_threshold=10, edge_window=21, plot=True, overwrite=True, **kwargs):
     """
     Normalize spectral flat field for long-slit observations. Parameters are optimized
     for NOT/ALFOSC spectra with horizontal slits, i.e., vertical spectra [axis=2],
@@ -245,32 +277,36 @@ def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, lower=0, u
     fname : string
         Input FITS file with raw lamp flat data
 
-    output : string [default='']
+    output : string  [default='']
         Filename of normalized flat frame, if not given the output is not saved to file
 
-    dispaxis : integer [default=2]
+    fig_dir : string  [default='']
+        Directory where the diagnostic plot is saved (if `plot=True`)
+
+    dispaxis : integer  [default=2]
         Dispersion axis, 1: horizontal spectra, 2: vertical spectra
 
-    lower : integer [default=0]
-        Mask pixels below this number in the fit to the spectral shape
+    overscan : integer  [default=50]
+        Overscan region, default for ALFOSC is 50 pixels on either side and on top
 
-    upper : integer [default=2050]
-        Mask pixels above this number in the fit to the spectral shape
+    order : integer  [default=24]
+        Order for Chebyshev polynomial to fit to the spatial profile (per row/col)
 
-    order : integer [default=24]
-        Order for Chebyshev polynomial to fit to the spectral shape.
+    savgol_window : integer  [default=51]
+        Window width in pixels for Savitzky--Golay filter of spatial profile
+
+    med_window : integer  [default=5]
+        Window width of median filter along spectral axis before fitting the spatial profile
+
+    edge_threshold : float  [default=10.]
+        The detection threshold for automatic edge detection
+
+    edge_window : integer  [default=21]
+        The Savitzky--Golay window used for automatic edge detection
 
     plot : boolean [default=True]
         Plot the 1d and 2d data for inspection?
 
-    show : boolean [default=True]
-        Show the figures directly or just save to file? If False, the figures will only be saved
-        as pdf files.
-
-    ext : integer [default=1]
-        File extension to open, default is 1 for ALFOSC which has a Primary extension with no data
-        and the Image extension containing the raw data.
-
     overwrite : boolean [default=False]
         Overwrite existing output file if True.
 
@@ -284,56 +320,74 @@ def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, lower=0, u
 
     """
     msg = list()
-
-    with pf.open(fname) as HDU:
-        if len(HDU)+1 <= ext:
-            flat = HDU[ext].data
-        else:
-            ext = 0
-            flat = HDU[0].data
-
-        if ext > 0 and HDU[0].size == 0:
-            # No data in first extension, merge headers:
-            hdr = HDU[0].header
-            for key in HDU[1].header.keys():
-                hdr[key] = HDU[1].header[key]
-
-        else:
-            hdr = HDU[ext].header
+    flat = pf.getdata(fname)
+    hdr = alfosc.get_alfosc_header(fname)
 
     msg.append("          - Input file: %s" % fname)
 
-    flat1D = np.mean(flat, dispaxis-1)
-    x = np.arange(len(flat1D))
-    lower = int(lower)
+    flat, hdr = trim_overscan(flat, hdr, overscan)
+    msg.append("          - Trimmed overscan: %i" % overscan)
+    # Get raw pixel array of spatial axis
+    x = np.arange(flat.shape[dispaxis-1])
+
+    # Detect edges of flat:
+    try:
+        edges = detect_flat_edges(flat, dispaxis=dispaxis,
+                                  savgol_window=edge_window, threshold=edge_threshold)
+        msg.append("          - Automatic edge detection found %i edges" % len(edges))
+        if len(edges) != 2:
+            msg.append("[WARNING] - Automatic edge detection failed. Using full frame!")
+            edges = [0, len(x)]
+    except:
+        msg.append("[WARNING] - Automatic edge detection failed. Using full frame!")
+        edges = [0, len(x)]
+
+    x1 = edges[0] + 5
+    x2 = edges[1] - 5
+
+    model = flat.copy()
     if dispaxis == 2:
-        upper = int(upper / hdr['DETYBIN'])
+        smoothing_kernel = (med_window, 1)
     else:
-        upper = int(upper / hdr['DETXBIN'])
-    fit = Chebyshev.fit(x[lower:upper], flat1D[lower:upper], order)
+        smoothing_kernel = (1, med_window)
+    smoothed_flat = ndimage.median_filter(flat, smoothing_kernel)
+    if dispaxis == 1:
+        # Flip image to iterate over rows:
+        smoothed_flat = smoothed_flat.T
+        model = model.T
+    msg.append("          - Median filtering flat frame along spectral axis to reduce noise")
+    msg.append("          - Fitting each spatial row/column using Chebyshev polynomials combined with Savitzky--Golay filtering")
+    msg.append("          - Polynomial order: %i" % order)
+    msg.append("          - Savitzky--Golay filter width: %i" % savgol_window)
+    pad = savgol_window // 2
+    x_fit = x[x1:x2]
+    for num, row in enumerate(smoothed_flat):
+        filtered_row = signal.savgol_filter(row[x1:x2], savgol_window, 1)
+        # Exclude filter edges, half filter width:
+        x_mask = x_fit[pad:-pad]
+        row_mask = filtered_row[pad:-pad]
+        fit_row = np.polynomial.Chebyshev.fit(x_mask, row_mask, order, domain=[x_fit.min(), x_fit.max()])
+        model1d = fit_row(x_fit)
+        # Remove edge effects in fitting, the filtered data are more robust:
+        # This stiched approach introduces a tiny discontinuity, but usually << 1%, so not important!
+        model1d[:pad] = filtered_row[:pad]
+        model1d[-pad:] = filtered_row[-pad:]
+        # Insert back into model image:
+        model[num][x1:x2] = model1d
 
-    flat_model = gaussian_filter1d(flat1D, sigma)
-    msg.append("          - Creating combined spectral model using Gaussian smoothing")
-    msg.append("          - and Chebyshev polynomium of degree: %i" % order)
+    if dispaxis == 1:
+        # Flip image the model back to original orientation:
+        model = model.T
 
-    # substitute the fit in the ends to remove convolution effects:
-    dx = len(x)-upper
-    ycut = len(x) - 2*dx
-    flat_model[:3*sigma] = fit(x[:3*sigma])
-    flat_model[ycut:] = fit(x[ycut:])
-
-    # make 2D spectral shape:
-    if dispaxis == 2:
-        model2D = np.resize(flat_model, flat.T.shape)
-        model2D = model2D.T
-    else:
-        model2D = np.resize(flat_model, flat.shape)
-
-    flat_norm = flat / model2D
+    flat_norm = flat / model
     hdr['DATAMIN'] = np.min(flat_norm)
     hdr['DATAMAX'] = np.max(flat_norm)
-    noise = np.std(flat1D - flat_model)
-    data_range = (np.min(flat_norm), np.max(flat_norm), np.median(flat_norm))
+    if dispaxis == 1:
+        stat_region = flat_norm[x1:x2, :]
+    else:
+        stat_region = flat_norm[:, x1:x2]
+    noise = np.std(stat_region)
+    data_range = (np.min(stat_region), np.max(stat_region), np.median(stat_region))
     msg.append("          - Standard deviation of 1D residuals: %.2f ADUs" % noise)
     msg.append("          - Normalized data range: min=%.2e  max=%.2e  median=%.2e" % data_range)
 
@@ -346,42 +400,55 @@ def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, lower=0, u
         ax2_2d = fig2D.add_subplot(122)
         ax1_2d.imshow(flat, origin='lower')
         ax1_2d.set_title("Raw Flat")
-        std_norm = np.std(flat_norm[lower:upper, :])
-        v1 = np.mean(flat_norm[lower:upper, :]) - 3*std_norm
-        v2 = np.mean(flat_norm[lower:upper, :]) + 3*std_norm
+        v1 = data_range[2] - 3*noise
+        v2 = data_range[2] + 3*noise
         ax2_2d.imshow(flat_norm, origin='lower', vmin=v1, vmax=v2)
         ax2_2d.set_title("Normalized Flat")
+        ax2_2d.set_yticklabels("")
         if dispaxis == 2:
-            ax1_2d.set_xlabel("Spatial Axis [pixels]")
-            ax2_2d.set_xlabel("Spatial Axis [pixels]")
-            ax1_2d.set_ylabel("Dispersion Axis [pixels]")
+            ax1_2d.set_xlabel("Spatial Axis [pixels]", fontsize=11)
+            ax2_2d.set_xlabel("Spatial Axis [pixels]", fontsize=11)
+            ax1_2d.set_ylabel("Dispersion Axis [pixels]", fontsize=11)
         else:
-            ax1_2d.set_ylabel("Spatial Axis [pixels]")
-            ax1_2d.set_xlabel("Dispersion Axis [pixels]")
-            ax2_2d.set_xlabel("Dispersion Axis [pixels]")
+            ax1_2d.set_ylabel("Spatial Axis [pixels]", fontsize=11)
+            ax1_2d.set_xlabel("Dispersion Axis [pixels]", fontsize=11)
+            ax2_2d.set_xlabel("Dispersion Axis [pixels]", fontsize=11)
 
+        # Plot 1D cross-section along slit:
         ax1_1d = fig1D.add_subplot(211)
         ax2_1d = fig1D.add_subplot(212)
 
-        residuals = flat1D - flat_model
-        ax1_1d.plot(x, flat1D, 'k.')
-        ax1_1d.plot(x, flat_model, 'crimson', lw=2, alpha=0.8)
-        ax2_1d.plot(x, residuals, 'crimson', lw=2, alpha=0.8)
-        ax2_1d.axhline(0., ls='--', color='k', lw=0.5)
+        flat1D = np.nanmedian(flat, 2-dispaxis)
+        f1d = signal.savgol_filter(flat1D[x1:x2], savgol_window, 1)
+        x_mask = x_fit[pad:-pad]
+        row_mask = f1d[pad:-pad]
+        fit_row = np.polynomial.Chebyshev.fit(x_mask, row_mask, order, domain=[x_fit.min(), x_fit.max()])
+        flat_model = fit_row(x_fit)
+        # Remove edge effects in fitting, the filtered data are more robust:
+        # This stiched approach introduces a tiny discontinuity, but usually << 1%, so not important!
+        flat_model[:pad] = f1d[:pad]
+        flat_model[-pad:] = f1d[-pad:]
 
-        ax2_1d.set_xlabel("Dispersion Axis [pixels]")
+        residuals = flat1D[x1:x2] - flat_model
+        ax1_1d.plot(x, flat1D, 'k-', lw=0.9)
+        ax1_1d.plot(x_fit, flat_model, 'crimson', lw=1.5, alpha=0.8)
+        ax2_1d.plot(x_fit, residuals/flat_model, 'k', lw=0.5)
+        ax2_1d.axhline(0., ls='--', color='0.3', lw=0.5)
+        ax1_1d.axvline(x1, color='b', ls=':', alpha=0.8)
+        ax1_1d.axvline(x2, color='b', ls=':', alpha=0.8)
+
+        ax2_1d.set_xlabel("Spatial Axis [pixels]", fontsize=11)
 
         power = np.floor(np.log10(np.max(flat1D))) - 1
         majFormatter = ticker.FuncFormatter(lambda x, p: my_formatter(x, p, power))
         ax1_1d.get_yaxis().set_major_formatter(majFormatter)
-        ax1_1d.set_ylabel('Counts  [$10^{{{0:d}}}$ ADU]'.format(int(power)))
+        ax1_1d.set_ylabel('Counts  [$10^{{{0:d}}}$ ADU]'.format(int(power)), fontsize=11)
 
-        power2 = np.floor(np.log10(np.max(residuals))) - 1
-        majFormatter2 = ticker.FuncFormatter(lambda x, p: my_formatter(x, p, power2))
-        ax2_1d.get_yaxis().set_major_formatter(majFormatter2)
-        ax2_1d.set_ylabel('Residual  [$10^{{{0:d}}}$ ADU]'.format(int(power2)))
-        noise = np.std(residuals[lower:upper])
-        ax2_1d.set_ylim(-8*noise, 8*noise)
+        ax2_1d.set_ylabel('($F_{\\rm 1D} -$ model) / model', fontsize=11)
+        plot_noise = np.nanstd(residuals/flat_model)
+        ax2_1d.set_ylim(-5*plot_noise, 5*plot_noise)
+        ax2_1d.set_xlim(x.min(), x.max())
+        ax1_1d.set_xlim(x.min(), x.max())
 
         ax1_1d.minorticks_on()
         ax2_1d.minorticks_on()
@@ -392,14 +459,13 @@ def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, lower=0, u
         fname_root = file_base.strip('.fits')
         fig1d_fname = os.path.join(fig_dir, "specflat_1d_%s.pdf" % fname_root)
         fig2d_fname = os.path.join(fig_dir, "specflat_2d_%s.pdf" % fname_root)
+        fig1D.tight_layout()
+        fig2D.tight_layout()
         fig1D.savefig(fig1d_fname)
         fig2D.savefig(fig2d_fname)
         msg.append("          - Saved graphic output for 1D model: %s" % fig1d_fname)
         msg.append("          - Saved graphic output for 2D model: %s" % fig2d_fname)
-        if show:
-            plt.show(block=True)
-        else:
-            plt.close('all')
+        plt.close('all')
 
     hdr['ORDER'] = (order, 'Order used for Chebyshev polynomial fit')
     hdr['NORMRMS'] = (noise, 'RMS noise of normalization [ADUs]')
@@ -420,80 +486,5 @@ def normalize_spectral_flat(fname, output='', fig_dir='', dispaxis=2, lower=0, u
     msg.append(" [OUTPUT] - Saving normalized MASTER FLAT: %s" % output)
     msg.append("")
     output_msg = "\n".join(msg)
-    if verbose:
-        print(output_msg)
 
     return output, output_msg
-
-
-if __name__ == '__main__':
-    parser = ArgumentParser()
-    parser.add_argument("--bias", type=str, nargs='+',
-                        help="Raw Bias frame(s)")
-    parser.add_argument("--bias-kappa", type=int, default=15,
-                        help="Threshold for sigma-kappa clipping in BIAS combiniation")
-    parser.add_argument("--flat", type=str, nargs='+',
-                        help="Raw Spectral flat frame(s)")
-    parser.add_argument("--flat-kappa", type=int, default=5,
-                        help="Threshold for sigma-kappa clipping in FLAT combiniation")
-    parser.add_argument("--flat-lower", type=int, default=0,
-                        help="Lower boundary on pixels used for spectral shape fitting")
-    parser.add_argument("--flat-upper", type=int, default=2050,
-                        help="Upper boundary on pixels used for spectral shape fitting")
-    parser.add_argument("--flat-slit", type=str, default='',
-                        help="Only combine flats taking with the given slit")
-    parser.add_argument("--flat-order", type=int, default=24,
-                        help="Polynomial order for fit to spectral shape")
-    parser.add_argument("--flat-sigma", type=int, default=5,
-                        help="Kernel width for Gaussian smoothing")
-    parser.add_argument("--flat-axis", type=int, default=2,
-                        help="Dispersion axis, 1: horizontal, 2: vertical")
-    parser.add_argument("--plot-dir", type=str, default="",
-                        help="Directory to save plots")
-    parser.add_argument("-p", "--plot", action="store_true",
-                        help="Plot diagnostics for spectral flat fielding?")
-    parser.add_argument("-s", "--show", action="store_true",
-                        help="Show diagnostics for spectral flat fielding?")
-    parser.add_argument("-x", "--ext", type=int, default=1,
-                        help="Extension number of input data")
-    parser.add_argument("-v", "--verbose", action="store_true",
-                        help="Print status updates")
-    args = parser.parse_args()
-
-    # Check if bias frames are present:
-    # ---------------------------------
-    if args.bias is not None:
-        if len(args.bias) == 1:
-            bias_frames = np.loadtxt(args.bias[0], usecols=(0,), dtype=str)
-
-        elif len(args.bias) > 1:
-            bias_frames = args.bias
-
-        else:
-            raise ValueError("Invalid input for --bias")
-
-        combine_bias_frames(bias_frames, output='MASTER_BIAS.fits',
-                            kappa=args.bias_kappa,
-                            verbose=args.verbose)
-
-    # Check if flat frames are present:
-    # ---------------------------------
-    if args.flat is not None:
-        if len(args.flat) == 1:
-            flat_frames = np.loadtxt(args.flat[0], usecols=(0,), dtype=str)
-
-        elif len(args.flat) > 1:
-            flat_frames = args.flat
-
-        else:
-            raise ValueError("Invalid input for --flat")
-
-        mflat_fname = combine_flat_frames(flat_frames, mbias='MASTER_BIAS.fits',
-                                          match_slit=args.flat_slit,
-                                          kappa=args.flat_kappa, verbose=args.verbose)
-
-        normalize_spectral_flat(mflat_fname, axis=args.flat_axis,
-                                x1=args.flat_x1, x2=args.flat_x2,
-                                order=args.flat_order, sigma=args.flat_sigma,
-                                plot=args.plot, show=args.show, ext=args.ext,
-                                overwrite=False, verbose=args.verbose)
