@@ -56,7 +56,7 @@ class Report(object):
 
     def commit(self, text):
         if self.verbose:
-            print(text)
+            print(text, end='', flush=True)
         self.lines.append(text)
 
     def error(self, text):
@@ -146,7 +146,6 @@ def run_pipeline(options_fname, object_id=None, verbose=False, interactive=False
             options[section_name].update(section)
         else:
             options[section_name] = section
-    raw_path = user_options['path']
 
     if object_id is None:
         pass
@@ -166,7 +165,6 @@ def run_pipeline(options_fname, object_id=None, verbose=False, interactive=False
         log.fatal_error()
         return
 
-
     if interactive:
         # Set all interactive steps to True
         options['identify']['interactive'] = True
@@ -174,37 +172,24 @@ def run_pipeline(options_fname, object_id=None, verbose=False, interactive=False
         options['extract']['interactive'] = True
         options['response']['interactive'] = True
 
-    # if raw_path is not None and not os.path.exists(raw_path):
-    #     log.error("Data path does not exist : %s" % raw_path)
-    #     log.fatal_error()
-    #     return
-
     dataset_fname = options['dataset']
-    if os.path.exists(dataset_fname):
+    if dataset_fname and os.path.exists(dataset_fname):
         # -- load collection
         database = io.load_database(dataset_fname)
         log.write("Loaded file classification database: %s" % dataset_fname)
         # -- reclassify (takes already identified files into account)
 
     else:
-        # Classify files:
-        log.write("Classyfying files in folder: %r" % raw_path)
-        try:
-            database, message = do.classify(raw_path, progress=verbose)
-            io.save_database(database, dataset_fname)
-            log.commit(message)
-            log.write("Saved file classification database: %s" % dataset_fname)
-        except ValueError as err:
-            log.error(str(err))
-            print(err)
-            log.fatal_error()
-            return
-        except FileNotFoundError as err:
-            log.error(str(err))
-            log.fatal_error()
-            return
+        log.error("Dataset does not exist : %s" % dataset_fname)
+        log.fatal_error()
+        return
 
     # -- Organize object files in dataset:
+    if 'SPEC_OBJECT' not in database:
+        log.error("No spectroscopic data found in the dataset!")
+        log.error("Check the classification table... object type 'SPEC_OBJECT' missing")
+        log.fatal_error()
+        return
     object_filelist = database['SPEC_OBJECT']
     object_images = list(map(do.RawImage, object_filelist))
 
@@ -232,6 +217,12 @@ def run_pipeline(options_fname, object_id=None, verbose=False, interactive=False
         if arc_type in database.keys():
             arc_images += database[arc_type]
 
+    if len(arc_images) == 0:
+        log.error("No arc line calibration data found in the dataset!")
+        log.error("Check the classification table... object type 'ARC_HeNe' or 'ARC_ThAr' missing")
+        log.fatal_error()
+        return
+
     arc_images_for_grism = defaultdict(list)
     for arc_img in arc_images:
         raw_grism = fits.getheader(arc_img)['ALGRNM']
@@ -316,13 +307,15 @@ def run_pipeline(options_fname, object_id=None, verbose=False, interactive=False
         # Loop over all:
         objects_to_reduce = object_images
     else:
-        object_names = [img.target_name for img in object_images]
         objects_to_reduce = list()
-        for obj_id in object_id:
-            if obj_id in object_names:
-                index = object_names.index(obj_id)
-                obj = object_images[index]
-                objects_to_reduce.append(obj)
+        for img in object_images:
+            if img.object in object_id:
+                objects_to_reduce.append(img)
+
+        if len(objects_to_reduce) == 0:
+            log.error("No data matched the given object ID: %r" % object_id)
+            log.fatal_error()
+            return
 
     for sci_img in objects_to_reduce:
         # Create working directory:
@@ -355,7 +348,7 @@ def run_pipeline(options_fname, object_id=None, verbose=False, interactive=False
         extract_pdf_fname = os.path.join(output_dir, 'extract1D_details.pdf')
 
         # Combine Bias Frames matched for CCD setup:
-        bias_frames = sci_img.match_files(database['BIAS'])
+        bias_frames = sci_img.match_files(database['BIAS'], date=False)
         if options['mbias']:
             master_bias_fname = options['mbias']
             log.write("Using static master bias frame: %s" % options['mbias'])
@@ -380,9 +373,14 @@ def run_pipeline(options_fname, object_id=None, verbose=False, interactive=False
 
 
         # Combine Flat Frames matched for CCD setup, grism, slit and filter:
-        flat_frames = sci_img.match_files(database['SPEC_FLAT'], grism=True, slit=True, filter=True)
+        flat_frames = sci_img.match_files(database['SPEC_FLAT'], date=False, grism=True, slit=True, filter=True)
         if options['mflat']:
-            norm_flat_fname = options['mflat']
+            if options['mflat'] is None:
+                norm_flat_fname = ''
+            elif options['mflat'].lower() in ['none', 'null']:
+                norm_flat_fname = ''
+            else:
+                norm_flat_fname = options['mflat']
             log.write("Using static master flat frame: %s" % options['mflat'])
         elif len(flat_frames) == 0:
             log.error("No flat frames provided!")
@@ -412,10 +410,7 @@ def run_pipeline(options_fname, object_id=None, verbose=False, interactive=False
                 log.write("Running task: Spectral Flat Normalization")
                 _, norm_msg = normalize_spectral_flat(comb_flat_fname, output=norm_flat_fname,
                                                       fig_dir=output_dir, dispaxis=sci_img.dispaxis,
-                                                      lower=options['flat']['lower'], upper=options['flat']['upper'],
-                                                      order=options['flat']['order'], sigma=options['flat']['sigma'],
-                                                      plot=options['flat']['plot'], show=options['flat']['show'],
-                                                      overwrite=True)
+                                                      **options['flat'])
                 log.commit(norm_msg)
                 log.add_linebreak()
                 status['master_flat'] = norm_flat_fname
@@ -456,9 +451,11 @@ def run_pipeline(options_fname, object_id=None, verbose=False, interactive=False
 
         # Response Function:
         if 'SPEC_FLUX-STD' in database:
-            flux_std_files = sci_img.match_files(database['SPEC_FLUX-STD'], grism=True, slit=True, filter=True, get_closest_time=True)
+            flux_std_files = sci_img.match_files(database['SPEC_FLUX-STD'],
+                                                 date=False, grism=True, slit=True, filter=True, get_closest_time=True)
         else:
             flux_std_files = []
+
         if len(flux_std_files) == 0:
             log.warn("No spectroscopic standard star was found in the dataset!")
             log.warn("The reduced spectra will not be flux calibrated")
@@ -466,7 +463,8 @@ def run_pipeline(options_fname, object_id=None, verbose=False, interactive=False
 
         else:
             std_fname = flux_std_files[0]
-            response_fname = os.path.join(output_dir, 'response_%s.fits' % (grism))
+            # response_fname = os.path.join(output_dir, 'response_%s.fits' % (grism))
+            response_fname = 'response_%s.fits' % (grism)
             if os.path.exists(response_fname) and not options['response']['force']:
                 log.write("Response function already exists: %s" % response_fname)
                 log.add_linebreak()
