@@ -40,7 +40,7 @@ import numpy as np
 import matplotlib.pyplot as plt
 from astropy.io import fits
 from scipy.ndimage import median_filter
-from scipy.signal import find_peaks
+from scipy import signal
 from numpy.polynomial import Chebyshev
 import os
 import warnings
@@ -54,10 +54,15 @@ from pynot.functions import mad, get_version_number
 __version__ = get_version_number()
 
 
-def trim_overscan(img, hdr, overscan=50):
+def trim_overscan(img, hdr, overscan=50, mode='spec'):
+    """Trim the overscan regions on either side in X and on top in Y"""
     # Trim overscan
-    X = create_pixel_array(hdr, 1)
-    Y = create_pixel_array(hdr, 2)
+    if mode == 'spec':
+        X = create_pixel_array(hdr, 1)
+        Y = create_pixel_array(hdr, 2)
+    else:
+        X = np.arange(img.shape[1]) + 1
+        Y = np.arange(img.shape[0]) + 1
     img_region_x = (X >= overscan) & (X <= 2148-overscan)
     xlimits = img_region_x.nonzero()[0]
     img_region_y = (Y <= 2102-overscan)
@@ -67,11 +72,94 @@ def trim_overscan(img, hdr, overscan=50):
     y1 = min(ylimits)
     y2 = max(ylimits)+1
     img_trim = img[y1:y2, x1:x2]
-    hdr['CRPIX1'] += x1
-    hdr['CRPIX2'] += y1
+    if mode == 'spec':
+        hdr['CRVAL1'] += x1
+        hdr['CRVAL2'] += y1
+
+    else:
+        hdr['CRPIX1'] -= x1
+        hdr['CRPIX2'] -= y1
+    hdr['NAXIS1'] = img_trim.shape[1]
+    hdr['NAXIS2'] = img_trim.shape[0]
     return img_trim, hdr
 
 
+def detect_filter_edge(fname, overscan=50):
+    """Automatically detect edges in the normalized flat field"""
+    # Get median profile along slit:
+    img = fits.getdata(fname)
+    hdr = get_alfosc_header(fname)
+    if 'OVERSCAN' in hdr:
+        overscan = 0
+
+    # Using the normalized flat field, the values are between 0 and 1.
+    # Convert the image to a binary mask image:
+    img[img > 0.5] = 1.
+    img[img < 0.5] = -1.
+
+    fx = np.median(img, 0) > 0.5
+    fy = np.median(img, 1) > 0.5
+
+    # Detect initial edges:
+    x1 = np.min(fx.nonzero()[0])
+    x2 = np.max(fx.nonzero()[0])
+
+    y1 = np.min(fy.nonzero()[0])
+    y2 = np.max(fy.nonzero()[0])
+
+    # If the edge is curved, decrease the trim edges
+    # until they are fully inside the image region
+    lower = img[y1, x1]
+    upper = img[y2, x2]
+    while lower < 0 and upper < 0:
+        x1 += 1
+        y1 += 1
+        x2 -= 1
+        y2 -= 1
+        lower = img[y1, x1]
+        upper = img[y2, x2]
+
+    return (x1-overscan, x2-overscan, y1, y2)
+
+
+def trim_filter_edge(fname, x1, x2, y1, y2, output='', output_dir=''):
+    """Trim image edges"""
+    # Get median profile along slit:
+    msg = list()
+    msg.append("          - Loaded file: %s" % fname)
+
+    if output == '':
+        basename = os.path.basename(fname)
+        output = 'trim_' + basename
+
+    if output_dir != '':
+        if not os.path.exists(output_dir):
+            os.mkdir(output_dir)
+        output = os.path.join(output_dir, output)
+
+    with fits.open(fname) as hdu_list:
+        for hdu in hdu_list:
+            if hdu.data is None:
+                continue
+
+            data = hdu.data
+            hdr = hdu.header
+            data_trim = data[y1:y2, x1:x2]
+            hdr['CRPIX1'] -= x1
+            hdr['NAXIS1'] = data_trim.shape[1]
+            hdr['CRPIX2'] -= y1
+            hdr['NAXIS2'] = data_trim.shape[0]
+            hdr.add_comment("Image trimmed by PyNOT")
+            hdu.data = data_trim
+            hdu.header = hdr
+        hdu_list.writeto(output, overwrite=True)
+    msg.append(" [OUTPUT] - Saving trimmed image: %s" % output)
+    msg.append("")
+    output_msg = "\n".join(msg)
+
+    return output_msg
+
+
 def fit_background_image(data, order_bg=3, xmin=0, xmax=None, kappa=10, fwhm_scale=3):
     """
     Fit background in 2D spectral data. The background is fitted along the spatial rows by a Chebyshev polynomium.
@@ -106,7 +194,7 @@ def fit_background_image(data, order_bg=3, xmin=0, xmax=None, kappa=10, fwhm_sca
         xmax = len(x) + xmax
     SPSF = np.nanmedian(data, 0)
     noise = 1.5*mad(SPSF)
-    peaks, properties = find_peaks(SPSF, prominence=kappa*noise, width=3)
+    peaks, properties = signal.find_peaks(SPSF, prominence=kappa*noise, width=3)
     mask = (x >= xmin) & (x <= xmax)
     for num, center in enumerate(peaks):
         width = properties['widths'][num]
@@ -172,7 +260,7 @@ def auto_fit_background(data_fname, output_fname, dispaxis=2, order_bg=3, kappa=
     """
     msg = list()
     data = fits.getdata(data_fname)
-    hdr = fits.getheader(data_fname)
+    hdr = get_alfosc_header(data_fname)
     if 'DISPAXIS' in hdr:
         dispaxis = hdr['DISPAXIS']
 
@@ -257,6 +345,7 @@ def correct_cosmics(input_fname, output_fname, niter=4, gain=None, readnoise=Non
     msg.append("          - Cosmic Ray Rejection using Astroscrappy (based on van Dokkum 2001)")
     sci = fits.getdata(input_fname)
     hdr = fits.getheader(input_fname)
+    hdr['EXTNAME'] = 'DATA'
     msg.append("          - Loaded input image: %s" % input_fname)
     with fits.open(input_fname) as hdu:
         if 'SKY' in hdu:
@@ -318,7 +407,7 @@ def correct_cosmics(input_fname, output_fname, niter=4, gain=None, readnoise=Non
     return output_msg
 
 
-def raw_correction(sci_raw, hdr, bias_fname, flat_fname, output='', overscan=50, overwrite=True):
+def raw_correction(sci_raw, hdr, bias_fname, flat_fname='', output='', overscan=50, overwrite=True, mode='spec'):
     """
     Perform bias subtraction, flat field correction, and cosmic ray rejection
 
@@ -334,8 +423,8 @@ def raw_correction(sci_raw, hdr, bias_fname, flat_fname, output='', overscan=50,
     bias_fname : string
         Filename of bias image to subtract from `sci_raw`
 
-    flat_fname : string
-        Filename of normalized flat field image
+    flat_fname : string  [default='']
+        Filename of normalized flat field image. If none is given, no flat field correction will be applied
 
     output : string  [default='']
         Output filename
@@ -354,109 +443,26 @@ def raw_correction(sci_raw, hdr, bias_fname, flat_fname, output='', overscan=50,
     """
     msg = list()
     mbias = fits.getdata(bias_fname)
-    msg.append("          - Loaded bias image: %s" % bias_fname)
-    mflat = fits.getdata(flat_fname)
-    mflat[mflat == 0] = 1
-    msg.append("          - Loaded flat field image: %s" % flat_fname)
-
-    sci = (sci_raw - mbias)/mflat
-
-    # Trim overscan
-    sci, hdr = trim_overscan(sci, hdr, overscan=overscan)
-
-    # Calculate error image:
-    if hdr['CCDNAME'] == 'CCD14':
-        hdr['GAIN'] = 0.16
-    gain = hdr['GAIN']
-    readnoise = hdr['RDNOISE']
-    with warnings.catch_warnings():
-        warnings.simplefilter('ignore')
-        err = np.sqrt(gain*sci + readnoise**2) / gain
-    msg.append("          - Created noise image")
-    msg.append("          - Gain=%.2f  and Read Noise=%.2f" % (gain, readnoise))
-
-    # Fix NaN values from negative pixel values:
-    err_NaN = np.isnan(err)
-    err[err_NaN] = readnoise/gain
-    msg.append("          - Correcting NaNs in noise image: %i pixel(s)" % np.sum(err_NaN))
-    hdr['DATAMIN'] = np.nanmin(sci)
-    hdr['DATAMAX'] = np.nanmax(sci)
-    hdr['EXTNAME'] = 'DATA'
-
-    mask = np.zeros_like(sci, dtype=int)
-    msg.append("          - Empty pixel mask created")
-    mask_hdr = fits.Header()
-    mask_hdr.add_comment("0 = Good Pixels")
-    mask_hdr.add_comment("1 = Cosmic Ray Hits")
-
-    sci_ext = fits.PrimaryHDU(sci, header=hdr)
-    err_ext = fits.ImageHDU(err, header=hdr, name='ERR')
-    mask_ext = fits.ImageHDU(mask, header=mask_hdr, name='MASK')
-    output_HDU = fits.HDUList([sci_ext, err_ext, mask_ext])
-    output_HDU.writeto(output, overwrite=overwrite)
-    msg.append("          - Successfully corrected the image.")
-    msg.append(" [OUTPUT] - Saving output: %s" % output)
-    msg.append("")
-    output_msg = "\n".join(msg)
-
-    return output_msg
-
-
-
-def correct_raw_file(input_fname, *, output, bias_fname, flat_fname='', overscan=50, overwrite=True):
-    """
-    Perform bias subtraction and flat field correction
-    NOTE: Add trimming parameters!!
-
-    Parameters
-    ==========
-
-    input_fname : string
-        Input filename of raw ALFOSC image
-
-    output : string
-        Output filename
-
-    bias_fname : string
-        Filename of bias image to subtract from `sci_raw`
-
-    flat_fname : string
-        Filename of normalized flat field image,
-        If not given, no flat field correction is performed
-
-
-    overscan : int  [default=50]
-        Number of pixels in overscan at the edge of the CCD.
-        The overscan region will be trimmed.
-
-    overwrite : boolean  [default=True]
-        Overwrite existing output file if True.
-
-    Returns
-    -------
-    output_msg : string
-        Log of status messages
-    """
-
-    msg = list()
-
-    hdr = get_alfosc_header(input_fname)
-    sci_raw = fits.getdata(input_fname)
-    msg.append("          - Loaded input image: %s" % input_fname)
-    mbias = fits.getdata(bias_fname)
+    bias_hdr = get_alfosc_header(bias_fname)
     msg.append("          - Loaded bias image: %s" % bias_fname)
     if flat_fname:
         mflat = fits.getdata(flat_fname)
         mflat[mflat == 0] = 1
+        flat_hdr = get_alfosc_header(flat_fname)
         msg.append("          - Loaded flat field image: %s" % flat_fname)
     else:
         mflat = 1.
         msg.append("          - Not flat field image provided. No correction applied!")
 
-    sci = (sci_raw - mbias)/mflat
+    # Trim overscan of raw image:
+    # - Trimming again of processed images doesn't change anything,
+    # - so do it just in case the input has not been trimmed
+    sci_raw, hdr = trim_overscan(sci_raw, hdr, overscan=overscan, mode=mode)
+    mbias, bias_hdr = trim_overscan(mbias, bias_hdr, overscan=overscan, mode=mode)
+    mflat, flat_hdr = trim_overscan(mflat, flat_hdr, overscan=overscan, mode=mode)
 
-    # Trim overscan
-    sci, hdr = trim_overscan(sci, hdr, overscan=overscan)
+    # Correct image:
+    sci = (sci_raw - mbias)/mflat
 
     # Calculate error image:
     if hdr['CCDNAME'] == 'CCD14':
@@ -477,12 +483,24 @@ def correct_raw_file(input_fname, *, output, bias_fname, flat_fname='', overscan
     hdr['DATAMAX'] = np.nanmax(sci)
     hdr['EXTNAME'] = 'DATA'
     hdr['AUTHOR'] = 'PyNOT version %s' % __version__
+    hdr['OVERSCAN'] = 'TRIMMED'
 
     mask = np.zeros_like(sci, dtype=int)
     msg.append("          - Empty pixel mask created")
     mask_hdr = fits.Header()
     mask_hdr.add_comment("0 = Good Pixels")
     mask_hdr.add_comment("1 = Cosmic Ray Hits")
+    for key in ['CRPIX1', 'CRPIX2', 'CRVAL1', 'CRVAL2', 'CTYPE1', 'CTYPE2', 'CUNIT1', 'CUNIT2']:
+        mask_hdr[key] = hdr[key]
+
+    if mode == 'spec':
+        mask_hdr['CDELT1'] = hdr['CDELT1']
+        mask_hdr['CDELT2'] = hdr['CDELT2']
+    else:
+        mask_hdr['CD1_1'] = hdr['CD1_1']
+        mask_hdr['CD1_2'] = hdr['CD1_2']
+        mask_hdr['CD2_1'] = hdr['CD2_1']
+        mask_hdr['CD2_2'] = hdr['CD2_2']
 
     sci_ext = fits.PrimaryHDU(sci, header=hdr)
     err_ext = fits.ImageHDU(err, header=hdr, name='ERR')
@@ -495,3 +513,24 @@ def correct_raw_file(input_fname, *, output, bias_fname, flat_fname='', overscan
     output_msg = "\n".join(msg)
 
     return output_msg
+
+
+
+def correct_raw_file(input_fname, *, output, bias_fname, flat_fname='', overscan=50, overwrite=True, mode='spec'):
+    """
+    Wrapper for `raw_correction` using file input instead of image input
+
+    Returns
+    -------
+    output_msg : string
+        Log of status messages
+    """
+    hdr = get_alfosc_header(input_fname)
+    sci_raw = fits.getdata(input_fname)
+    msg = "          - Loaded input image: %s" % input_fname
+
+    output_msg = raw_correction(sci_raw, hdr, bias_fname, flat_fname, output=output,
+                                overscan=overscan, overwrite=overwrite, mode=mode)
+    output_msg = msg + '\n' + output_msg
+
+    return output_msg
